#include "xmame.h"
#include <sys/stat.h>
#include <unistd.h>
#include <dirent.h>
#ifdef ZLIB
#include <zlib.h>
#endif

typedef enum
{
	kPlainFile,
	kZippedFile,
	kGZippedFile
} eFileType;

typedef struct
{
	FILE		*file;
	unsigned char	*data;
	int		offset;
	int		length;
 	eFileType	type;
} FakeFileHandle;

int load_zipped_file (const char *zipfile,const char *filename,
	unsigned char **buf, int *length);

static char *rompathv[MAXPATHC];
static int   rompathc = 0;

/* helper function which decomposes a path list into a vector of paths */
void init_rom_path(void)
{
	char *token;
	/* has to be static since ptr's to it are kept after this function */
	static char buf[MAXPATHL];
#ifdef MESS
	/* mess also searches the current dir */
	static char  *current_dir=".";
	rompathv[0] = current_dir;
	rompathc    = 1;
#endif
	
	/* make a copy since strtok modifies rompath */
	strcpy(buf,rompath);
	token = strtok(buf, ":");
	while ((rompathc < MAXPATHC) && token)
	{
		rompathv[rompathc] = token;
		rompathc++;
		token = strtok (NULL, ":");
	}
}

/*
 * check if roms/samples for a game exist at all
 * return 1 on success, otherwise 0
 */
int osd_faccess(const char *filename, int filetype)
{
#ifndef MESS
	char name[MAXPATHL]; /* FIXME buffer overrun !! */
	int i;
	
	switch (filetype)
	{
		case OSD_FILETYPE_ROM:
		case OSD_FILETYPE_SAMPLE:
		    for(i=0;i<rompathc;i++)
		    {
			sprintf(name,"%s/%s",rompathv[i],filename);
			if (access(name, F_OK) == 0) return 1;
			/* try with a .zip extension */
			if  (filetype == OSD_FILETYPE_ROM)
				sprintf(name,"%s/%s/%s.zip",rompathv[i],"roms", filename);
			else
				sprintf(name,"%s/%s/%s.zip",rompathv[i],"samples", filename);
			if (access(name, F_OK) == 0) return 1;
		    }
		    break;
	}
	
	return 0;
#else
	return 1; /* dirty, but know way to tell if it's there since we don't 
		     have mandatory dirs with mess */
#endif
}

/*
 * Search file caseinsensitively
 *
 * Arguments: 
 *	char * path - complete pathname to the desired file. The string will
 *	              be modified during search (and contains the final output).
 *
 * Return TRUE if found, FALSE otherwise.
 */
int filesearch(char *path)
{
    DIR *dirp;
    struct dirent *de = NULL;
    char *ep, *dp, *fp;

    ep = strrchr(path, '/');
    if (ep) {
	*ep = '\0';	/* I guess root directory is not supported */
	dp = path;
	fp = ep + 1;
    } else {
	dp = "."; /* well, what should be the correct name for "current dir" */
	fp = path;
    }

    if (*fp == '\0') {
	return FALSE;
    }

    /* jamc: code to perform a upper/lowercase rom search */
    /* try to open directory */
    if ((dirp = opendir(dp)) == (DIR *) 0) { 
	return FALSE;
    }

    /* search entry and upcasecompare to desired file name */
    for (de = readdir(dirp); de; de = readdir(dirp))
	if (!strcasecmp(de->d_name, fp)) break;
    if (de) strcpy(fp, de->d_name);
    closedir(dirp);
    
    if (ep) *ep = '/';

    if (de) return TRUE;
    return FALSE;
}

/*
 * file handling routines
 *
 * gamename holds the driver name, filename is only used for ROMs and samples.
 * if 'write' is not 0, the file is opened for write. Otherwise it is opened
 * for read.
 */
void *osd_fopen(const char *gamename, const char *filename, int filetype, 
 int write) {
	char *pt, name[256]; /* FIXME ?? possible buffer overrun */
	FakeFileHandle *f;
	int i;

	f = (FakeFileHandle *)malloc(sizeof(FakeFileHandle));
	if (f == NULL) return f;
	
	f->type = kPlainFile;
	f->file = 0;

	switch (filetype)
	{
#ifdef MESS
		case OSD_FILETYPE_ROM_CART:
		case OSD_FILETYPE_IMAGE:
#else
		case OSD_FILETYPE_ROM:
		case OSD_FILETYPE_SAMPLE:
#endif
		    if (write)
		    {
			if (errorlog)
			    fprintf(errorlog, "Error trying to open rom/sample %s in write mode\n", filename);
			free(f);
			return NULL;
		    }
		    
		    for(i=0;i<rompathc;i++)
		    {
#ifdef MESS
			sprintf(name,"%s/%s",rompathv[i],filename);
			if (filesearch(name))
			{
				f->file = fopen(name, "rb");
				if (f->file) return f;
			}
#endif
			sprintf(name,"%s/%s/%s",rompathv[i],gamename,filename);
			if (filesearch(name))
			{
				f->file = fopen(name, "rb");
				if (f->file) return f;
			}
			/* try with a .zip extension */
#ifdef MESS
			/* first try rompath/filename.zip */
			sprintf(name,"%s/%s",rompathv[i],filename);
			if ( (pt=strrchr(name, '.')) ) *pt = 0;
			strcat(name,".zip");
			f->file = fopen(name, "rb");
			if (f->file == 0)
			{
			    /* failed now try rompath/systemname/filename.zip */
			    sprintf(name,"%s/%s/%s",rompathv[i],gamename,filename);
			    if ( (pt=strrchr(name, '.')) ) *pt = 0;
			    strcat(name,".zip");
			    f->file = fopen(name, "rb");
			}
#else
			if  (filetype == OSD_FILETYPE_ROM)
				sprintf(name,"%s/%s/%s.zip",rompathv[i],"roms", gamename);
			else
				sprintf(name,"%s/%s/%s.zip",rompathv[i],"samples", gamename);
			f->file = fopen(name, "rb");
#endif
			if (f->file)
			{
				if (errorlog)
					fprintf(errorlog, "using zip file for %s\n", filename);
				fclose(f->file);
				if (load_zipped_file(name, filename, &f->data, &f->length) == 0)
				{
					f->type   = kZippedFile;
					f->offset = 0;
					return f;
				}
				else
					f->file   = 0;
			}
#ifdef ZLIB
			/* try with a .gz extension */
#ifdef MESS
			sprintf(name,"%s/%s.gz",rompathv[i],filename);
			if ( (f->file = gzopen(name, "rb")) )
			{
				f->type=kGZippedFile;
				return f;
			}
#endif
			sprintf(name,"%s/%s/%s.gz",rompathv[i],gamename,filename);
			if ( (f->file = gzopen(name, "rb")) )
			{
				f->type=kGZippedFile;
				return f;
			}
#endif
		    }
		    break;
		case OSD_FILETYPE_HIGHSCORE:
		    if (mame_highscore_enabled())
		    {
			sprintf(name,"%s/%s.hi", spooldir, gamename);
			f->file = fopen(name,write ? "wb" : "rb");
		    }
		    break;
		case OSD_FILETYPE_CONFIG:
		    pt = getenv("HOME");
		    if (pt)
		    {
			sprintf(name, "%s/%s/%s.cfg", pt, NAME, gamename);
			f->file = fopen(name,write ? "wb" : "rb");
		    }
		    break;
		case OSD_FILETYPE_INPUTLOG:
			pt = getenv("HOME");
			if (pt)
			{
				sprintf(name,"%s/%s.inp",pt,filename);
				f->file = fopen(name,write ? "wb" : "rb");
			}
			break;
	}

	if (f->file == 0)
	{
		free(f); 
		return NULL;
	}

	return f;
}



int osd_fread(void *file,void *buffer,int length)
{
	FakeFileHandle *f = (FakeFileHandle *)file;

	switch (f->type)
	{
		case kPlainFile:
			return fread(buffer,1,length,f->file);
			break;
		case kZippedFile:
			/* reading from the uncompressed image of a zipped file */
			if (f->data)
			{
				if (length + f->offset > f->length)
					length = f->length - f->offset;
				memcpy(buffer, f->offset + f->data, length);
				f->offset += length;
				return length;
			}
			break;
#ifdef ZLIB
		case kGZippedFile:
			return gzread(f->file, buffer, length);
			break;
#endif			
	}

	return 0;
}



int osd_fwrite(void *file,const void *buffer,int length)
{
	return fwrite(buffer,1,length,((FakeFileHandle *)file)->file);
}



int osd_fseek(void *file,int offset,int whence)
{
	FakeFileHandle *f = (FakeFileHandle *)file;

	switch (f->type)
	{
		case kPlainFile:
			return fseek(((FakeFileHandle *)file)->file,offset,whence);
			break;
		case kZippedFile:
			/* seeking within the uncompressed image of a zipped file */
			switch (whence)
			{
				case SEEK_SET:
					f->offset = offset;
					break;
				case SEEK_CUR:
					f->offset += offset;
					break;
				case SEEK_END:
					f->offset = f->length + offset;
					break;
			}
			break;
#ifdef ZLIB
		case kGZippedFile: /* jugh fseek is not nice on gzip files */
			if (whence != SEEK_END)
				return gzseek(f->file, offset, whence);
			else /* bummer zlib doesn't support SEEK_END so lett's be creative */
			{
				char buf[256];
				/* okay first of all lett's find the end of this file */
				while (gzread(f->file, buf, 256) == 256) {}
				return gzseek(f->file, offset, SEEK_CUR);
			}
			break;
#endif
	}

	return -1;
}



void osd_fclose(void *file)
{
	FakeFileHandle *f = (FakeFileHandle *) file;

	switch(f->type)
	{
		case kPlainFile:
			fclose(f->file);
			break;
		case kZippedFile:
			if (f->data)
				free(f->data);
			break;
#ifdef ZLIB
		case kGZippedFile:
			gzclose(f->file);
			break;
#endif
	}
	free(f);
}
