/*
 * X-Mame keyboard specifics code
 *
 */
#define __KEYBOARD_C_

/*
 * Include files.
 */
#include "xmame.h"
 
#ifndef ggi

#ifdef svgalib
#include <vgakeyboard.h>
#define sysdep_update_keyboard keyboard_update
#endif

/*
 * Wait until a key is pressed and return its code.
 */
int osd_read_key(void)
{
	int i;

	/* wait for all keys to be released */
	do
	{
	      sysdep_update_keyboard();
	      for (i = OSD_MAX_KEY;i >= 0; i--)
	          if (key[i]) break;
	} while (i >= 0);

	/* wait for a key press */
	for(;;)
	{
	      sysdep_update_keyboard();
	      for (i = OSD_MAX_KEY;i >= 0;i--)
	          if (key[i])
	          {
#ifdef svgalib
	             if (i==SCANCODE_RIGHTCONTROL) i = OSD_KEY_RCONTROL;
	             if (i==SCANCODE_RIGHTALT)     i = OSD_KEY_ALTGR;
#endif
	             return i;
	          }
	}
}

int osd_read_keyrepeat() 
{
	int i;
	memset((void *)&key[0],FALSE,OSD_MAX_KEY*sizeof(byte) );
	/* wait for a key press */
	for(;;)
	{
	      sysdep_update_keyboard();
	      for (i = OSD_MAX_KEY;i >= 0;i--)
	          if (key[i])
	          {
#ifdef svgalib
	             if (i==SCANCODE_RIGHTCONTROL) i = OSD_KEY_RCONTROL;
	             if (i==SCANCODE_RIGHTALT)     i = OSD_KEY_ALTGR;
#endif
#ifdef MESS
	             return key_to_pseudo_code(i);
#else
	             return i;
#endif
	          }
	}
}

/* check if a key is pressed. The keycode is the standard PC keyboard code, as */
/* defined in osdepend.h. Return 0 if the key is not pressed, nonzero otherwise. */
int osd_key_pressed(int keycode)
{
   /* blames to the dos-people who want to check key states before
      the display (and under X thus the keyboard) is initialised */
   if (!key) return 0;

   sysdep_update_keyboard();

   if (keycode > OSD_MAX_KEY)
   {
	switch (keycode)
	{
		case OSD_KEY_FAST_EXIT:
		case OSD_KEY_UNPAUSE:
			return key[OSD_KEY_ESC];

		case OSD_KEY_RESET_MACHINE:
			return key[OSD_KEY_F3];

		case OSD_KEY_VOLUME_DOWN:
			return (key[OSD_KEY_MINUS_PAD] && !key[OSD_KEY_LSHIFT]);

		case OSD_KEY_VOLUME_UP:
			return (key[OSD_KEY_PLUS_PAD] && !key[OSD_KEY_LSHIFT]);

		case OSD_KEY_PAUSE:
			return key[OSD_KEY_P];

		case OSD_KEY_CONFIGURE:
			return key[OSD_KEY_TAB];

		case OSD_KEY_SHOW_GFX:
			return key[OSD_KEY_F4];
#ifdef MESS			
		case OSD_KEY_FRAMESKIP:
			return key[OSD_KEY_F8];
		case OSD_KEY_THROTTLE:
			return key[OSD_KEY_F10];
		case OSD_KEY_SHOW_FPS:
			return key[OSD_KEY_F11];
		case OSD_KEY_SNAPSHOT:
			return key[OSD_KEY_F12];
		case OSD_KEY_GAMMA_DOWN:
			return (key[OSD_KEY_MINUS_PAD] && key[OSD_KEY_LSHIFT]);
		case OSD_KEY_GAMMA_UP:
			return (key[OSD_KEY_PLUS_PAD] && key[OSD_KEY_LSHIFT]);
#endif
	}
   }
   
#ifdef svgalib
   if (keycode == OSD_KEY_RCONTROL) keycode = SCANCODE_RIGHTCONTROL;
   if (keycode == OSD_KEY_ALTGR) keycode = SCANCODE_RIGHTALT;
#endif

   return key[keycode];
}

#endif /* #ifndef ggi */

/* return the name of a key */
const char *osd_key_name(int keycode)
{
  static char *keynames[89] = { 
    "None", "ESC", "1", "2", "3", "4", "5", "6", "7", "8", "9", "0",
    "MINUS", "EQUAL", "BACKSPACE", "TAB", "Q", "W", "E", "R", "T", "Y",
    "U", "I", "O", "P", "OPENBRACE", "CLOSEBRACE", "ENTER", "LEFTCONTROL",
    "A", "S", "D", "F", "G", "H", "J", "K", "L", "COLON", "QUOTE",
    "TILDE", "LEFTSHIFT", "Unknow", "Z", "X", "C", "V", "B", "N", "M", "COMMA",
    "STOP", "SLASH", "RIGHTSHIFT", "ASTERISK", "ALT", "SPACE", "CAPSLOCK",
    "F1", "F2", "F3", "F4", "F5", "F6", "F7", "F8", "F9", "F10", "NUMLOCK",
    "SCRLOCK", "HOME", "UP", "PGUP", "MINUS PAD", "LEFT", "5 PAD", "RIGHT",
    "PLUS PAD", "END", "DOWN", "PGDN", "INS", "DEL", "RIGHTCONTROL", "ALTGR",
    "Unknown", "F11", "F12" 
  };
  static char *unknown = "Unknown";

  if (keycode < 89) return (char*)keynames[keycode];
  return (char *)unknown;
}

/* dummy functions we will probably never be able to blink the keybleds 
   under unix anyway */
void osd_led_w(int led,int on) 
{
	return;
}

#ifdef MESS
int key_to_pseudo_code(int k)
{
    switch (k)
    {
        case OSD_KEY_F3:
            return OSD_KEY_RESET_MACHINE;
        case OSD_KEY_F4:
            return OSD_KEY_SHOW_GFX;
        case OSD_KEY_F8:
            return OSD_KEY_FRAMESKIP;
        case OSD_KEY_F10:
            return OSD_KEY_THROTTLE;
        case OSD_KEY_F11:
            return OSD_KEY_SHOW_FPS;
        case OSD_KEY_F12:
	    return OSD_KEY_SNAPSHOT;
	case OSD_KEY_TAB:
            return OSD_KEY_CONFIGURE;
	case OSD_KEY_ESC:
	    return OSD_KEY_FAST_EXIT;
        case OSD_KEY_MINUS_PAD:
            if (key[OSD_KEY_LSHIFT])
                return OSD_KEY_GAMMA_DOWN;
            return OSD_KEY_VOLUME_DOWN;
        case OSD_KEY_PLUS_PAD:
            if (key[OSD_KEY_LSHIFT])
                return OSD_KEY_GAMMA_UP;
            return OSD_KEY_VOLUME_UP;
        case OSD_KEY_P:
            return OSD_KEY_PAUSE;
    }
    return k;
}
#endif
