/*
 * X-mame main-routine
 */

#define __MAIN_C_
#include "xmame.h"
#include "sound.h"
#include "driver.h"

/* put here anything you need to do when the program is started. Return 0 if */
/* initialization was successful, nonzero otherwise. */
int osd_init(void)
{
	/* now invoice system-dependent initialization */
	sysdep_keyboard_init();
	sysdep_mouse_init();
	sysdep_joy_initvars();
	sysdep_audio_initvars();
	
#ifdef USE_TIMER
#ifndef svgalib
	/* in svgalib mode must start timer AFTER open display */
	if (play_sound)
	   if (start_timer()==OSD_NOT_OK) return(OSD_NOT_OK);
#endif
#endif
	return OSD_OK;
}

/*
 * Cleanup routines to be executed when the program is terminated.
 */
void osd_exit (void)
{
	sysdep_keyboard_close();
	sysdep_mouse_close();
	sysdep_joy_close();
	sysdep_exit();
}


int main (int argc, char **argv)
{
	int res;
	unsigned char lsb_test[2]={0,1};

#ifdef svgalib
        /* first let's try to init svgalib */
        /* and give up those evil root rights, */
        /* esp since the rest of the code is full of bufferoveruns ;( */
	vga_init();
#endif
#ifdef ggi
	/* ggi needs to be initialized too */
     	ggiInit();
#endif
#ifdef xf86_dga
	if(xf86_init() == OSD_NOT_OK) return OSD_NOT_OK;
#endif


       /* Lett's see of the endians of this arch is correct otherwise
          YELL about it and bail out. */
#ifdef LSB_FIRST
	if(*((unsigned short*)lsb_test) != 0x0100)
#else	
	if(*((unsigned short*)lsb_test) != 0x0001)
#endif
	{
	   fprintf(stderr, "Error: Compiled byte ordering doesn't match machine byte ordering\n"
	      "Are you sure you choose the right arch?\n"
#ifdef LSB_FIRST
	      "Compiled for lsb-first try removing -DLSB_FIRST from DEFS.$(ARCH)\nin makefile.unix\n");
#else
	      "Compiled for msb-first try adding -DLSB_FIRST to DEFS.$(ARCH)\nin makefile.unix\n");
#endif
	   return OSD_NOT_OK;
	}
	
        /* Set the title, now auto build from defines from the makefile */
        sprintf(title,"%s (%s) version %s", NAME, DISPLAY_METHOD, VERSION);

	/* parse configuration file and environment */
	if (get_config() == OSD_NOT_OK) return OSD_NOT_OK;
	
	/* handle -rompath before checking the frontend options */
	for (res=1; res < argc; res++)
	{
	   if (strcasecmp(argv[res], "-rompath") == 0)
	   {
	      res++;
	      if (res==argc)
	      {
	         fprintf(stderr, "-rompath requires an argument\n");
	         return OSD_NOT_OK;
	      }
	      strncat(rompath, ":",       (MAXPATHL-1) - strlen(rompath));
 	      strncat(rompath, argv[res], (MAXPATHL-1) - strlen(rompath));
	   }
	}
	
	init_rom_path();

	/* check for frontend options */
	if ( (res = frontend_help (argc, argv)) != 1234 ) return res;
	
	/* parse the cmdline */
	if (parse_command_line(argc, argv) == OSD_NOT_OK) return (OSD_NOT_OK);

	/* this could change the samplerate so has to be done here and not
	   in osd_init ! */
	sysdep_init();

	/* Set the samplerate to 0 if sound is disabled,
	   this should notify mame not to "render" it (speed improvement) */
	if (play_sound==0) 
	{
	   if (fake_sound)
	      options.samplerate=8000;
	   else
	   {
	      options.samplerate=0;
	      No_FM=0;
	   }
	}
	
	/* go for it */
	res = run_game (game_index , &options);

	/* close open files */
	if (options.errorlog) fclose (options.errorlog);
	if (options.playback) osd_fclose (options.playback);
	if (options.record)   osd_fclose (options.record);

	return res;
}
